// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "config.h"

#include "id3reader.h"

#include <glibmm/stringutils.h>
#include <iostream>

bool ID3TagReader::mpg123_inited = false;

ID3TagReader::ID3TagReader(const std::string& filename)
	: m_FileName(filename),
	m_HaveID3(false),
	m_NumChannels(0),
	m_Duration(0)
{
	if (!mpg123_inited)
	{
		mpg123_init();
		mpg123_inited = true;
	}

	m_Handle = mpg123_new(NULL, NULL);
}

ID3TagReader::~ID3TagReader()
{
	mpg123_delete(m_Handle);
}

ID3TagReader::ID3File::ID3File(mpg123_handle* handle, const std::string& filename)
	: m_Handle(handle)
{
	if (!(m_IsValid = mpg123_open(m_Handle, filename.c_str()) == MPG123_OK))
		std::cerr << "Cannot open " << filename <<
			" (" << mpg123_strerror(m_Handle) << ")" << std::endl;
}

ID3TagReader::ID3File::~ID3File()
{
	if (m_IsValid)
		mpg123_close(m_Handle);
}

bool ID3TagReader::Load()
{
	ID3File id3_file(m_Handle, m_FileName);
	if (id3_file)
	{
		mpg123_scan(m_Handle);
		m_HaveID3 = (mpg123_meta_check(m_Handle) & MPG123_ID3) &&
			mpg123_id3(m_Handle, &m_ID3v1, &m_ID3v2) == MPG123_OK;

		// Compute the track duration and number of channels now while the
		// file is open.
		uint32_t length = (uint32_t)mpg123_length(m_Handle);
		long rate = 0;
		int channels, encoding;
		mpg123_getformat(m_Handle, &rate, &channels, &encoding);
		m_Duration = (uint32_t)((float) length * (float) 1000 / (float) rate);
		m_NumChannels = channels;
	}

	return m_HaveID3;
}

Glib::ustring ID3TagReader::GetTitle()
{
	return (m_HaveID3 && m_ID3v2->title) ? m_ID3v2->title->p : "";
}

Glib::ustring ID3TagReader::GetArtist()
{
	return (m_HaveID3 && m_ID3v2->artist) ? m_ID3v2->artist->p : "";
}

Glib::ustring ID3TagReader::GetAlbum()
{
	return (m_HaveID3 && m_ID3v2->album) ? m_ID3v2->album->p : "";
}

Glib::ustring ID3TagReader::GetGenre()
{
	const Glib::ustring genre_table[] =
	{
		"Blues",
		"Classic Rock",
		"Country",
		"Dance",
		"Disco",
		"Funk",
		"Grunge",
		"Hip-Hop",
		"Jazz",
		"Metal",
		"New Age",
		"Oldies",
		"Other",
		"Pop",
		"R&B",
		"Rap",
		"Reggae",
		"Rock",
		"Techno",
		"Industrial",
		"Alternative",
		"Ska",
		"Death Metal",
		"Pranks",
		"Soundtrack",
		"Euro-Techno",
		"Ambient",
		"Trip-Hop",
		"Vocal",
		"Jazz+Funk",
		"Fusion",
		"Trance",
		"Classical",
		"Instrumental",
		"Acid",
		"House",
		"Game",
		"Sound Clip",
		"Gospel",
		"Noise",
		"AlternRock",
		"Bass",
		"Soul",
		"Punk",
		"Space",
		"Meditative",
		"Instrumental Pop",
		"Instrumental Rock",
		"Ethnic",
		"Gothic",
		"Darkwave",
		"Techno-Industrial",
		"Electronic",
		"Pop-Folk",
		"Eurodance",
		"Dream",
		"Southern Rock",
		"Comedy",
		"Cult",
		"Gangsta",
		"Top 40",
		"Christian Rap",
		"Pop/Funk",
		"Jungle",
		"Native American",
		"Cabaret",
		"New Wave",
		"Psychedelic",
		"Rave",
		"Showtunes",
		"Trailer",
		"Lo-Fi",
		"Tribal",
		"Acid Punk",
		"Acid Jazz",
		"Polka",
		"Retro",
		"Musical",
		"Rock & Roll",
		"Hard Rock",
		"Folk",
		"Folk/Rock",
		"National folk",
		"Swing",
		"Fast-fusion",
		"Bebob",
		"Latin",
		"Revival",
		"Celtic",
		"Bluegrass",
		"Avantgarde",
		"Gothic Rock",
		"Progressive Rock",
		"Psychedelic Rock",
		"Symphonic Rock",
		"Slow Rock",
		"Big Band",
		"Chorus",
		"Easy Listening",
		"Acoustic",
		"Humour",
		"Speech",
		"Chanson",
		"Opera",
		"Chamber Music",
		"Sonata",
		"Symphony",
		"Booty Bass",
		"Primus",
		"Porn Groove",
		"Satire",
		"Slow Jam",
		"Club",
		"Tango",
		"Samba",
		"Folklore",
		"Ballad",
		"Powder Ballad",
		"Rhythmic Soul",
		"Freestyle",
		"Duet",
		"Punk Rock",
		"Drum Solo",
		"A Capella",
		"Euro-House",
		"Dance Hall",
		"Goa",
		"Drum & Bass",
		"Club House",
		"Hardcore",
		"Terror",
		"Indie",
		"BritPop",
		"NegerPunk",
		"Polsk Punk",
		"Beat",
		"Christian Gangsta",
		"Heavy Metal",
		"Black Metal",
		"Crossover",
		"Contemporary C",
		"Christian Rock",
		"Merengue",
		"Salsa",
		"Thrash Metal",
		"Anime",
		"JPop",
		"SynthPop"
	};

	if (!m_HaveID3 || !m_ID3v2->genre) return "";

	if (m_ID3v2->genre->p[0] == '(')
	{
		guint genre_index = atoi(m_ID3v2->genre->p + 1);
		return (genre_index < sizeof(genre_table) / sizeof(Glib::ustring)) ?
			genre_table[genre_index] : "";
	}
	else
		return m_ID3v2->genre->p;
}

Glib::ustring ID3TagReader::GetYear()
{
	if (!m_HaveID3) return "";

	if (m_ID3v2->year)
		return m_ID3v2->year->p;
	else
	{
		guint i = 0;
		for (; i < m_ID3v2->texts && !Glib::str_has_prefix(m_ID3v2->text[i].id, "TDRC"); ++i);
		return (i < m_ID3v2->texts) ? m_ID3v2->text[i].text.p : "";
	}
}

uint16_t ID3TagReader::GetTrackNumber()
{
	if (!m_HaveID3) return 0;

	guint i = 0;
	for(; i < m_ID3v2->texts && !Glib::str_has_prefix(m_ID3v2->text[i].id, "TRCK"); ++i);
	return (i < m_ID3v2->texts) ? atoi((&m_ID3v2->text[i].text)->p) : 0;
}

uint16_t ID3TagReader::GetNumChannels()
{
	return m_NumChannels;
}

uint32_t ID3TagReader::GetDuration()
{
	return m_Duration;
}
